"""Module that provides Qt-specific functionality for psygnal.

This module provides convenience functions for starting and stopping a QTimer that
will monitor "queued" signals and invoke their callbacks.  This is useful when
psygnal is used in a Qt application, and you'd like to emit signals from a thread
but have their callbacks invoked in the main thread.
"""

from __future__ import annotations

from threading import Thread, current_thread

from ._queue import emit_queued

try:
    from qtpy.QtCore import Qt, QTimer
except (ImportError, RuntimeError):  # pragma: no cover
    raise ImportError(
        "The psygnal.qt module requires qtpy and some Qt backend to be installed"
    ) from None

_TIMERS: dict[Thread, QTimer] = {}


def start_emitting_from_queue(
    msec: int = 0,
    timer_type: Qt.TimerType = Qt.TimerType.PreciseTimer,
    thread: Thread | None = None,
) -> None:
    """Start a QTimer that will monitor the global emission queue.

    If a QTimer is already running in the current thread, then this function will
    update the interval and timer type of that QTimer. (It is safe to call this
    function multiple times in the same thread.)

    When callbacks are connected to signals with `connect(type='queued')`, then they
    are not invoked immediately, but rather added to a global queue.  This function
    starts a QTimer that will periodically check the queue and invoke any callbacks
    that are waiting to be invoked (in whatever thread this QTimer is running in).

    Parameters
    ----------
    msec : int, optional
        The interval (in milliseconds) at which the QTimer will check the global
        emission queue.  By default, the QTimer will check the queue as often as
        possible (i.e. 0 milliseconds).
    timer_type : Qt.TimerType, optional
        The type of timer to use.  By default, Qt.PreciseTimer is used, which is
        the most accurate timer available on the system.
    thread : Thread, optional
        The thread in which to start the QTimer.  By default, the QTimer will be
        started in the thread from which this function is called.
    """
    _thread = current_thread() if thread is None else thread
    if _thread not in _TIMERS:
        _TIMERS[_thread] = QTimer()

        _TIMERS[_thread].timeout.connect(emit_queued)

    _TIMERS[_thread].setTimerType(timer_type)
    if _TIMERS[_thread].isActive():
        _TIMERS[_thread].setInterval(msec)
    else:
        _TIMERS[_thread].start(msec)


def stop_emitting_from_queue(thread: Thread | None = None) -> None:
    """Stop the QTimer that monitors the global emission queue.

    thread : Thread, optional
        The thread in which to stop the QTimer. By default, will stop any QTimers
        in the thread from which this function is called.
    """
    _thread = current_thread() if thread is None else thread
    if (timer := _TIMERS.get(_thread)) is not None:
        timer.stop()



"""Dict that emits events when altered."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping, MutableMapping, Sequence
from functools import partial
from typing import TYPE_CHECKING, Any, ClassVar, TypeAlias, TypeVar, get_args

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler, SerializationInfo
    from typing_extensions import Self

from psygnal._group import EmissionInfo, PathStep, SignalGroup
from psygnal._signal import Signal, SignalInstance

_K = TypeVar("_K")
_V = TypeVar("_V")
TypeOrSequenceOfTypes: TypeAlias = type[_V] | Sequence[type[_V]]
DictArg: TypeAlias = Mapping[_K, _V] | Iterable[tuple[_K, _V]]


class TypedMutableMapping(MutableMapping[_K, _V]):
    """Dictionary that enforces value type.

    Parameters
    ----------
    data : Union[Mapping[_K, _V], Iterable[Tuple[_K, _V]], None], optional
        Data suitable of passing to dict(). Mapping of {key: value} pairs, or
        Iterable of two-tuples [(key, value), ...], or None to create an
    basetype : TypeOrSequenceOfTypes, optional
        Type or Sequence of Type objects. If provided, values entered into this Mapping
        must be an instance of one of the provided types. by default ()
    """

    def __init__(
        self,
        data: DictArg | None = None,
        *,
        basetype: TypeOrSequenceOfTypes = (),
        **kwargs: _V,
    ):
        self._dict: dict[_K, _V] = {}
        self._basetypes: tuple[type[_V], ...] = (
            tuple(basetype) if isinstance(basetype, Sequence) else (basetype,)
        )
        self.update({} if data is None else data, **kwargs)

    def __setitem__(self, key: _K, value: _V) -> None:
        self._dict[key] = self._type_check(value)

    def __delitem__(self, key: _K) -> None:
        del self._dict[key]

    def __getitem__(self, key: _K) -> _V:
        return self._dict[key]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> Iterator[_K]:
        return iter(self._dict)

    def __repr__(self) -> str:
        return repr(self._dict)

    def _type_check(self, value: _V) -> _V:
        """Check the types of items if basetypes are set for the model."""
        if self._basetypes and not any(isinstance(value, t) for t in self._basetypes):
            raise TypeError(
                f"Cannot add object with type {type(value)} to TypedDict expecting"
                f"type {self._basetypes}"
            )
        return value

    def __newlike__(self, mapping: MutableMapping[_K, _V]) -> Self:
        new = self.__class__()
        # separating this allows subclasses to omit these from their `__init__`
        new._basetypes = self._basetypes
        new.update(mapping)
        return new

    def copy(self) -> Self:
        """Return a shallow copy of the dictionary."""
        return self.__newlike__(self)

    def __copy__(self) -> Self:
        return self.copy()

    # PYDANTIC SUPPORT

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> Mapping[str, Any]:
        """Return the Pydantic core schema for this object."""
        from pydantic_core import core_schema

        def _serialize(obj: EventedDict[_K, _V], info: SerializationInfo, /) -> Any:
            if info.mode_is_json():
                return obj._dict
            return cls(obj._dict)

        # get key/value types
        key_type = val_type = Any
        if args := get_args(source_type):
            key_type = args[0]
            if len(args) > 1:
                val_type = args[1]

        # get key/value schemas and validators
        keys_schema = handler.generate_schema(key_type)
        values_schema = handler.generate_schema(val_type)
        dict_schema = core_schema.dict_schema(
            keys_schema=keys_schema,
            values_schema=values_schema,
        )
        return core_schema.no_info_after_validator_function(
            function=cls,
            schema=dict_schema,
            json_schema_input_schema=dict_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                _serialize,
                info_arg=True,
            ),
        )


class DictSignalInstance(SignalInstance):
    def _psygnal_relocate_info_(self, emission_info: EmissionInfo) -> EmissionInfo:
        """Relocate the emission info to the key being modified.

        (All signals on EventedDict have the key as the first argument.)
        """
        if args := emission_info.args:
            return emission_info.insert_path(PathStep(key=args[0]))
        return emission_info


DictSignal = partial(Signal, signal_instance_class=DictSignalInstance)


class DictEvents(SignalGroup):
    """Events available on [EventedDict][psygnal.containers.EventedDict]."""

    adding = DictSignal(object)  # (key, )
    """`(key,)` emitted before an item is added at `key`"""
    added = DictSignal(object, object)  # (key, value)
    """`(key, value)` emitted after a `value` is added at `key`"""
    changing = DictSignal(object)  # (key, )
    """`(key, old_value, new_value)` emitted before `old_value` is replaced with
    `new_value` at `key`"""
    changed = DictSignal(object, object, object)  # (key, old_value, value)
    """`(key, old_value, new_value)` emitted before `old_value` is replaced with
    `new_value` at `key`"""
    removing = DictSignal(object)  # (key, )
    """`(key,)` emitted before an item is removed at `key`"""
    removed = DictSignal(object, object)  # (key, value)
    """`(key, value)` emitted after `value` is removed at `key`"""


class EventedDict(TypedMutableMapping[_K, _V]):
    """Mutable mapping that emits events when altered.

    This class is designed to behave exactly like the builtin [`dict`][], but
    will emit events before and after all mutations (addition, removal, and
    changing).

    Parameters
    ----------
    data : Union[Mapping[_K, _V], Iterable[Tuple[_K, _V]], None], optional
        Data suitable of passing to dict(). Mapping of {key: value} pairs, or
        Iterable of two-tuples [(key, value), ...], or None to create an
    basetype : TypeOrSequenceOfTypes, optional
        Type or Sequence of Type objects. If provided, values entered into this Mapping
        must be an instance of one of the provided types. by default ().

    Attributes
    ----------
    events: DictEvents
        The `SignalGroup` object that emits all events available on an `EventedDict`.
    """

    events: DictEvents  # pragma: no cover
    _psygnal_group_: ClassVar[str] = "events"

    def __init__(
        self,
        data: DictArg | None = None,
        *,
        basetype: TypeOrSequenceOfTypes = (),
        **kwargs: _V,
    ):
        self.events = DictEvents()
        super().__init__(data, basetype=basetype, **kwargs)

    def __setitem__(self, key: _K, value: _V) -> None:
        if key not in self._dict:
            self.events.adding.emit(key)
            super().__setitem__(key, value)
            self.events.added.emit(key, value)
        else:
            old_value = self._dict[key]
            if value is not old_value:
                self.events.changing.emit(key)
                super().__setitem__(key, value)
                self.events.changed.emit(key, old_value, value)

    def __delitem__(self, key: _K) -> None:
        item = self._dict[key]
        self.events.removing.emit(key)
        super().__delitem__(key)
        self.events.removed.emit(key, item)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"



from __future__ import annotations

import inspect
import os
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING

import psygnal

if TYPE_CHECKING:
    from collections.abc import Container, Sequence

    from ._signal import SignalInstance


ROOT = str(Path(psygnal.__file__).parent)


class EmitLoopError(Exception):
    """Error type raised when an exception occurs during a callback."""

    __module__ = "psygnal"

    def __init__(
        self,
        exc: BaseException,
        signal: SignalInstance | None = None,
        recursion_depth: int = 0,
        reemission: str | None = None,
        emit_queue: Sequence[tuple] = (),
    ) -> None:
        # if isinstance(exc, EmitLoopError):
        #     super().__init__("nested EmitLoopError.")
        #     return

        self.__cause__ = exc

        # grab the signal name or repr
        if signal is None:  # pragma: no cover
            sig_name: str = ""
        elif instance := signal.instance:
            inst_class = instance.__class__
            mod = getattr(inst_class, "__module__", "")
            if mod:
                mod += "."
            sig_name = f"{mod}{inst_class.__qualname__}.{signal.name}"
        else:
            sig_name = signal.name

        msg = _build_psygnal_exception_msg(sig_name, exc, recursion_depth)

        # queued emission can be confusing, because the `signal.emit()` call shown
        # in the traceback will not match the emission that actually raised the error.
        if reemission == "queued" and (depth := len(emit_queue) - 1):
            msg += (
                "\nNOTE: reemission is set to 'queued', and this error occurred "
                f"at a queue-depth of {depth}.\nEmitting arguments: {emit_queue[-1]})\n"
            )

        super().__init__(msg)


def _build_psygnal_exception_msg(
    sig_name: str, exc: BaseException, recursion_depth: int
) -> str:
    msg = f"\n\nWhile emitting signal {sig_name!r}, an error occurred in a callback:"
    line = f"\n\n  {exc.__class__.__name__}: {exc}"
    msg += line + "\n  " + "-" * (len(line) - 4)

    if recursion_depth:
        s = "s" if recursion_depth > 1 else ""
        msg += f"\nnested {recursion_depth} level{s} deep"

    # get the first frame in the stack that is not in the psygnal package
    stack = inspect.stack()
    with suppress(Exception):
        emit_frame = next(fi for fi in stack if ROOT not in fi.filename)
        msg += "\n\n  SIGNAL EMISSION: \n"
        with suppress(IndexError):
            back_frame = stack[stack.index(emit_frame) + 1]
            msg += f"    {_fmt_frame(back_frame)}\n"
        msg += f"    {_fmt_frame(emit_frame)}  # <-- SIGNAL WAS EMITTED HERE\n"

    # get the last frame in the traceback, the one that raised the exception
    if tb := exc.__traceback__:
        with suppress(Exception):
            if not (inner := inspect.getinnerframes(tb)):
                return msg  # pragma: no cover

            except_frame = inner[-1]

            # show the immediately connected callback first
            first_cb = next((fi for fi in inner if ROOT not in fi.filename), None)
            if first_cb and first_cb != except_frame:
                num_inner = len(inner) - inner.index(first_cb) - 2
                msg += "\n  CALLBACK CHAIN:\n"
                msg += f"    {_fmt_frame(first_cb, with_context=False)}"
                msg += f"    ... {num_inner} more frames ...\n"

            # Then end with the frame that raised the exception
            msg += f"    {_fmt_frame(except_frame)}  # <-- ERROR OCCURRED HERE \n"
            if flocals := except_frame.frame.f_locals:
                if not os.getenv("PSYGNAL_HIDE_LOCALS"):
                    msg += "\n      Local variables:\n"
                    msg += _fmt_locals(flocals)

    return msg


def _fmt_frame(fi: inspect.FrameInfo, with_context: bool = True) -> str:
    msg = f"{fi.filename}:{fi.lineno} in {fi.function}\n"
    if with_context and (code_ctx := fi.code_context):
        msg += f"      {code_ctx[0].strip()}"
    return msg


def _fmt_locals(
    f_locals: dict, exclude: Container[str] = ("self", "cls"), name_width: int = 20
) -> str:
    lines = []
    for name, value in f_locals.items():
        if name not in exclude:
            val_repr = repr(value)
            if len(val_repr) > 60:
                val_repr = val_repr[:60] + "..."  # pragma: no cover
            lines.append("{:>{}} = {}".format(name, name_width, val_repr))
    return "\n".join(lines)



"""Utilities for testing psygnal Signals."""

from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING
from unittest.mock import Mock
from unittest.util import safe_repr

from psygnal import SignalGroup, SignalInstance

if TYPE_CHECKING:
    from collections.abc import Iterator
    from threading import Thread
    from typing import Any, Literal

    from typing_extensions import Self, TypedDict

    class ConnectKwargs(TypedDict, total=False):
        """Kwargs for SignalInstance.connect."""

        thread: Thread | Literal["main", "current"] | None
        check_nargs: bool | None
        check_types: bool | None
        unique: bool | Literal["raise"]
        max_args: int | None
        on_ref_error: Literal["raise", "warn", "ignore"]
        priority: int


__all__ = [
    "SignalTester",
    "assert_emitted",
    "assert_emitted_once",
    "assert_emitted_once_with",
    "assert_emitted_with",
    "assert_ever_emitted_with",
    "assert_not_emitted",
]


class SignalTester:
    """A tester object that listens to a signal and records its emissions.

    This class wraps a [`SignalInstance`][psygnal.SignalInstance] and a
    [`unittest.mock.Mock`][] object. It provides methods to connect and disconnect the
    mock from the signal, and to assert that the signal was emitted with the expected
    arguments.  It also behaves as a **context manager**, so you can monitor emissions
    of a signal within a specific context.

    !!! important

        The signal is *not* automatically connected to the mock when the SignalTester is
        created. You must call [`connect()`][psygnal.testing.SignalTester.connect] or
        use the context manager to connect the mock to the signal.

    Parameters
    ----------
    signal : SignalInstance | SignalGroup
        The signal instance or group to test.
    connect_kwargs : ConnectKwargs
        Keyword arguments to pass to the
        [`SignalInstance.connect()`][psygnal.SignalInstance.connect] method.

    Attributes
    ----------
    signal_instance : SignalInstance
        The signal instance being tested. If a `SignalGroup` is passed, it uses the
        `_psygnal_relay` attribute to get the underlying `SignalInstance`.
    mock : unittest.mock.Mock
        The mock object that will be connected to the signal.

    Examples
    --------
    ```python
    from psygnal import Signal
    from psygnal.testing import SignalTester


    class MyObject:
        value_changed = Signal(int)


    obj = MyObject()
    tester = SignalTester(obj.value_changed)
    tester.assert_not_emitted()

    with tester:
        obj.value_changed.emit(1)

    tester.assert_emitted()
    tester.assert_emitted_once()
    tester.assert_emitted_once_with(1)
    assert tester.emit_count == 1
    tester.reset()
    assert tester.emit_count == 0
    ```
    """

    def __init__(
        self,
        signal: SignalInstance | SignalGroup,
        connect_kwargs: ConnectKwargs | None = None,
    ) -> None:
        super().__init__()
        self.mock = Mock()
        if isinstance(signal, SignalGroup):
            signal_instance: SignalInstance = signal._psygnal_relay
        else:
            signal_instance = signal
        self.signal_instance: SignalInstance = signal_instance
        self.connect_kwargs = connect_kwargs or {}

    def reset(self) -> None:
        """Reset the underlying mock object."""
        self.mock.reset_mock()

    @property
    def emit_count(self) -> int:
        """Return the number of times the signal was emitted."""
        return self.mock.call_count

    @property
    def emit_args(self) -> tuple[Any, ...]:
        """Return the arguments of the last emission of the signal."""
        if (call_args := self.mock.call_args) is None:
            return ()

        return call_args[0]  # type: ignore[no-any-return]

    @property
    def emit_args_list(self) -> list[tuple[Any, ...]]:
        """Return the arguments of all emissions of the signal."""
        return [call[0] for call in self.mock.call_args_list]

    def connect(self) -> None:
        """Connect the mock to the signal."""
        self.signal_instance.connect(self.mock, **self.connect_kwargs)

    def disconnect(self) -> None:
        """Disconnect the mock from the signal."""
        self.signal_instance.disconnect(self.mock)

    def __enter__(self) -> Self:
        """Connect the mock to the signal."""
        self.connect()
        return self

    def __exit__(self, *args: Any) -> None:
        """Disconnect the mock from the signal."""
        self.disconnect()

    @property
    def signal_name(self) -> str:
        """Return the name of the signal."""
        return self.signal_instance.name or "signal"

    def assert_not_emitted(self) -> None:
        """Assert that the signal was never emitted."""
        if self.mock.call_count != 0:
            if self.mock.call_count == 1:
                n = "once"
            else:
                n = f"{self.mock.call_count} times"
            raise AssertionError(
                f"Expected {self.signal_name!r} to not have been emitted. Emitted {n}."
            )

    def assert_emitted(self) -> None:
        """Assert that the signal was emitted at least once."""
        if self.mock.call_count == 0:
            raise AssertionError(f"Expected {self.signal_name!r} to have been emitted.")

    def assert_emitted_once(self) -> None:
        """Assert that the signal was emitted exactly once."""
        if not self.mock.call_count == 1:
            raise AssertionError(
                f"Expected {self.signal_name!r} to have been emitted once. "
                f"Emitted {self.mock.call_count} times."
            )

    def assert_emitted_with(self, /, *args: Any) -> None:
        """Assert that the *last* emission of the signal had the given arguments."""
        if self.mock.call_args is None:
            raise AssertionError(
                f"Expected {self.signal_name!r} to have been emitted with arguments "
                f"{args!r}.\nActual: not emitted"
            )

        actual = self.mock.call_args[0]
        if actual != args:
            raise AssertionError(
                f"Expected {self.signal_name!r} to have been emitted with arguments "
                f"{args!r}.\nActual: {actual}"
            )

    def assert_emitted_once_with(self, /, *args: Any) -> None:
        """Assert that the signal was emitted exactly once with the given arguments."""
        if not self.mock.call_count == 1:
            raise AssertionError(
                f"Expected {self.signal_name!r} to have been emitted exactly once. "
                f"Emitted {self.mock.call_count} times."
            )

        actual = self.mock.call_args[0]
        if actual != args:
            raise AssertionError(
                f"Expected {self.signal_name!r} to have been emitted once with "
                f"arguments {args!r}.\nActual: {safe_repr(actual)}"
            )

    def assert_ever_emitted_with(self, /, *args: Any) -> None:
        """Assert that the signal was emitted *ever* with the given arguments."""
        if self.mock.call_args is None:
            raise AssertionError(
                f"Expected {self.signal_name!r} to have been emitted at least once "
                f"with arguments {args!r}.\nActual: not emitted"
            )

        actual = [call[0] for call in self.mock.call_args_list]
        if not any(call == args for call in actual):
            _actual: tuple | list = actual[0] if len(actual) == 1 else actual
            raise AssertionError(
                f"Expected {self.signal_name!r} to have been emitted at least once "
                f"with arguments {args!r}.\nActual: {safe_repr(_actual)}"
            )


@contextmanager
def assert_emitted(
    signal: SignalInstance | SignalGroup, connect_kwargs: ConnectKwargs | None = None
) -> Iterator[SignalTester]:
    """Assert that a signal was emitted at least once.

    Parameters
    ----------
    signal : SignalInstance | SignalGroup
        The signal instance or group to test.
    connect_kwargs : ConnectKwargs
        Keyword arguments to pass to the
        [`SignalInstance.connect()`][psygnal.SignalInstance.connect] method.

    Raises
    ------
    AssertionError
        If the signal was never emitted.
    """
    with SignalTester(signal, connect_kwargs) as mock:
        yield mock
        mock.assert_emitted()


@contextmanager
def assert_emitted_once(
    signal: SignalInstance | SignalGroup, connect_kwargs: ConnectKwargs | None = None
) -> Iterator[SignalTester]:
    """Assert that a signal was emitted exactly once.

    Parameters
    ----------
    signal : SignalInstance | SignalGroup
        The signal instance or group to test.
    connect_kwargs : ConnectKwargs
        Keyword arguments to pass to the
        [`SignalInstance.connect()`][psygnal.SignalInstance.connect] method.

    Raises
    ------
    AssertionError
        If the signal was emitted more than once.
    """
    with SignalTester(signal, connect_kwargs) as mock:
        yield mock
        mock.assert_emitted_once()


@contextmanager
def assert_not_emitted(
    signal: SignalInstance | SignalGroup, connect_kwargs: ConnectKwargs | None = None
) -> Iterator[SignalTester]:
    """Assert that a signal was never emitted.

    Parameters
    ----------
    signal : SignalInstance | SignalGroup
        The signal instance or group to test.
    connect_kwargs : ConnectKwargs
        Keyword arguments to pass to the
        [`SignalInstance.connect()`][psygnal.SignalInstance.connect] method.

    Raises
    ------
    AssertionError
        If the signal was emitted at least once.
    """
    with SignalTester(signal, connect_kwargs) as mock:
        yield mock
        mock.assert_not_emitted()


@contextmanager
def assert_emitted_with(
    signal: SignalInstance | SignalGroup,
    *args: Any,
    connect_kwargs: ConnectKwargs | None = None,
) -> Iterator[SignalTester]:
    """Assert that the *last* emission of the signal had the given arguments.

    Parameters
    ----------
    signal : SignalInstance | SignalGroup
        The signal instance or group to test.
    *args : Any
        The arguments to check for in the last emission of the signal.
    connect_kwargs : ConnectKwargs
        Keyword arguments to pass to the
        [`SignalInstance.connect()`][psygnal.SignalInstance.connect] method.

    Raises
    ------
    AssertionError
        If the signal was never emitted or if the last emission did not have the
        expected arguments.
    """
    with assert_emitted(signal, connect_kwargs) as mock:
        yield mock
        mock.assert_emitted_with(*args)


@contextmanager
def assert_emitted_once_with(
    signal: SignalInstance | SignalGroup,
    *args: Any,
    connect_kwargs: ConnectKwargs | None = None,
) -> Iterator[SignalTester]:
    """Assert that the signal was emitted exactly once with the given arguments.

    Parameters
    ----------
    signal : SignalInstance | SignalGroup
        The signal instance or group to test.
    *args : Any
        The arguments to check for in the last emission of the signal.
    connect_kwargs : ConnectKwargs
        Keyword arguments to pass to the
        [`SignalInstance.connect()`][psygnal.SignalInstance.connect] method.

    Raises
    ------
    AssertionError
        If the signal was not emitted or was emitted more than once or if the last
        emission did not have the expected arguments.
    """
    with assert_emitted_once(signal, connect_kwargs) as mock:
        yield mock
        mock.assert_emitted_once_with(*args)


@contextmanager
def assert_ever_emitted_with(
    signal: SignalInstance | SignalGroup,
    *args: Any,
    connect_kwargs: ConnectKwargs | None = None,
) -> Iterator[SignalTester]:
    """Assert that the signal was emitted *ever* with the given arguments.

    Parameters
    ----------
    signal : SignalInstance | SignalGroup
        The signal instance or group to test.
    *args : Any
        The arguments to check for in any emission of the signal.
    connect_kwargs : ConnectKwargs
        Keyword arguments to pass to the
        [`SignalInstance.connect()`][psygnal.SignalInstance.connect] method.

    Raises
    ------
    AssertionError
        If the signal was never emitted or if it was emitted but not with the expected
        arguments.
    """
    with assert_emitted(signal, connect_kwargs) as mock:
        yield mock
        mock.assert_ever_emitted_with(*args)



from __future__ import annotations

import inspect
import sys
import warnings
import weakref
from functools import partial
from types import BuiltinMethodType, FunctionType, MethodType, MethodWrapperType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    Protocol,
    TypeVar,
    cast,
)
from warnings import warn

from ._async import get_async_backend
from ._mypyc import mypyc_attr

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import TypeAlias, TypeGuard

    import toolz

    from ._async import _AsyncBackend

    RefErrorChoice: TypeAlias = Literal["raise", "warn", "ignore"]

__all__ = ["WeakCallback", "weak_callback"]
_T = TypeVar("_T")
_R = TypeVar("_R")  # return type of cb


def _is_toolz_curry(obj: Any) -> TypeGuard[toolz.curry]:
    """Return True if obj is a toolz.curry object."""
    tz = sys.modules.get("toolz")
    return False if tz is None else isinstance(obj, tz.curry)


def weak_callback(
    cb: Callable[..., _R] | WeakCallback[_R],
    *args: Any,
    max_args: int | None = None,
    finalize: Callable[[WeakCallback], Any] | None = None,
    strong_func: bool = True,
    on_ref_error: RefErrorChoice = "warn",
    priority: int = 0,
) -> WeakCallback[_R]:
    """Create a weakly-referenced callback.

    This function creates a weakly-referenced callback, with special considerations
    for many known callable types (functions, lambdas, partials, bound methods,
    partials on bound methods, builtin methods, etc.).

    NOTE: For the sake of least-surprise, an exception is made for functions and,
    lambdas, which are strongly-referenced by default.  See the `strong_func` parameter
    for more details.

    Parameters
    ----------
    cb : callable
        The callable to be called.
    *args
        Additional positional arguments to be passed to the callback (similar
        to functools.partial).
    max_args : int, optional
        The maximum number of positional arguments to pass to the callback.
        If provided, additional arguments passed to WeakCallback.cb will be ignored.
    finalize : callable, optional
        A callable that will be called when the callback is garbage collected.
        The callable will be passed the WeakCallback instance as its only argument.
    strong_func : bool, optional
        If True (default), a strong reference will be kept to the function `cb` if
        it is a function or lambda.  If False, a weak reference will be kept.  The
        reasoning for this is that functions and lambdas are very often defined *only*
        to be passed to this function, and would likely be immediately garbage
        collected if we weakly referenced them. If you would specifically like to
        *allow* the function to be garbage collected, set this to False.
    on_ref_error : {'raise', 'warn', 'ignore'}, optional
        What to do if a weak reference cannot be created.  If 'raise', a
        ReferenceError will be raised.  If 'warn' (default), a warning will be issued
        and a strong-reference will be used. If 'ignore' a strong-reference will be
        used (silently).
    priority : int, optional
        The priority of the callback.  This is used to determine the order in which
        callbacks are called when multiple are connected to the same signal.
        Higher priority callbacks are called first. Negative values are allowed.
        The default is 0.

    Returns
    -------
    WeakCallback
        A WeakCallback subclass instance appropriate for the given callable.
        The fast way to "call" the callback is to use the `cb` method, passing a
        single args tuple, it returns nothing.  A `__call__` method is also provided,
        that can be used to call the original function as usual.

    Examples
    --------
    ```python
        from psygnal._weak_callback import weak_callback

    class T:
        def greet(self, name):
            print("hello,", name)

    def _on_delete(weak_cb):
        print("deleting!")

    t = T()
    weak_cb = weak_callback(t.greet, finalize=_on_delete)

    weak_cb.cb(("world",))  # "hello, world"

    del t  # "deleting!"

    weak_cb.cb(("world",))  # ReferenceError
    ```
    """
    if isinstance(cb, WeakCallback):
        return cb

    kwargs: dict[str, Any] | None = None
    if isinstance(cb, partial):
        args = cb.args + args
        kwargs = cb.keywords
        cb = cb.func

    is_coro = inspect.iscoroutinefunction(cb)
    if is_coro:
        if (backend := get_async_backend()) is None:
            raise RuntimeError(
                "Cannot create async callback yet... No async backend set. "
                "Please call `psygnal.set_async_backend()` before connecting."
            )
        if not backend.running.is_set():
            warnings.warn(
                f"\n\nConnection of async {cb.__name__!r} will not do anything!\n"
                "Async backend not running. Launch `get_async_backend().run()` "
                "in a background task and wait for `backend.running`",
                RuntimeWarning,
                stacklevel=2,
            )

    if isinstance(cb, FunctionType):
        # NOTE: I know it looks like this should be easy to express in much shorter
        # syntax ... but mypyc will likely fail at runtime.
        # Make sure to test compiled version if you change this.
        if strong_func:
            if is_coro:
                return StrongCoroutineFunction(
                    cb, max_args, args, kwargs, priority=priority
                )
            return StrongFunction(cb, max_args, args, kwargs, priority=priority)
        else:
            if is_coro:
                return WeakCoroutineFunction(
                    cb,
                    max_args,
                    args,
                    kwargs,
                    finalize,
                    on_ref_error=on_ref_error,
                    priority=priority,
                )
            return WeakFunction(
                cb, max_args, args, kwargs, finalize, on_ref_error, priority=priority
            )

    if isinstance(cb, MethodType):
        if getattr(cb, "__name__", None) == "__setitem__":
            try:
                key = args[0]
            except IndexError as e:  # pragma: no cover
                raise TypeError(
                    "WeakCallback.__setitem__ requires a key argument"
                ) from e
            obj = cast("SupportsSetitem", cb.__self__)
            return WeakSetitem(
                obj, key, max_args, finalize, on_ref_error, priority=priority
            )

        if is_coro:
            return WeakCoroutineMethod(
                cb, max_args, args, kwargs, finalize, on_ref_error, priority=priority
            )
        return WeakMethod(
            cb, max_args, args, kwargs, finalize, on_ref_error, priority=priority
        )

    if isinstance(cb, (MethodWrapperType, BuiltinMethodType)):
        if kwargs:  # pragma: no cover
            raise NotImplementedError(
                "MethodWrapperTypes do not support keyword arguments"
            )

        if cb is setattr:
            try:
                obj, attr = args[:2]
            except IndexError as e:  # pragma: no cover
                raise TypeError(
                    "setattr requires two arguments, an object and an attribute name."
                ) from e
            return WeakSetattr(
                obj, attr, max_args, finalize, on_ref_error, priority=priority
            )
        return WeakBuiltin(
            cb, max_args, args, finalize, on_ref_error, priority=priority
        )

    if _is_toolz_curry(cb):
        cb_partial = getattr(cb, "_partial", None)
        if cb_partial is None:  # pragma: no cover
            raise TypeError(
                "toolz.curry object found without a '_partial' attribute. This "
                "version of toolz is not supported. Please open an issue at psygnal."
            )
        return weak_callback(
            cb_partial,
            *args,
            max_args=max_args,
            finalize=finalize,
            on_ref_error=on_ref_error,
            priority=priority,
        )

    if callable(cb):
        # this is a bit of hack to workaround a segfault observed in testing
        # on python <=3.11 when compiled by mypyc,
        # during _weak_callback___WeakFunction_traverse
        # it specifically happens with MethodWrapperType objects, that I think are made
        # by mypyc itself.  So we just don't attempt to weakref them here anymore.
        _call = getattr(cb, "__call__", None)  # noqa
        if isinstance(_call, MethodWrapperType):
            return StrongFunction(_call, max_args, args, kwargs, priority=priority)

        return WeakFunction(
            cb, max_args, args, kwargs, finalize, on_ref_error, priority=priority
        )

    raise TypeError(f"unsupported type {type(cb)}")  # pragma: no cover


class WeakCallback(Generic[_R]):
    """Abstract Base Class for weakly-referenced callbacks.

    Do not instantiate this class directly, use the `weak_callback` function instead.
    The main public-facing methods of all subclasses are:

        cb(args: tuple[Any, ...] = ()) -> None: special fast callback method, args only.
        dereference() -> Callable[..., _R] | None: return strong dereferenced callback.
        __call__(*args: Any, **kwargs: Any) -> _R: call original callback
        __eq__: compare two WeakCallback instances for equality
        object_key: static method that returns a unique key for an object.

    NOTE: can't use ABC here because then mypyc and PySide2 don't play nice together.
    """

    def __init__(
        self,
        obj: Any,
        max_args: int | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
    ) -> None:
        self._key: str = WeakCallback.object_key(obj)
        self._obj_module: str = getattr(obj, "__module__", None) or ""
        self._obj_qualname: str = getattr(obj, "__qualname__", "")
        self._object_repr: str = WeakCallback.object_repr(obj)
        self._max_args: int | None = max_args
        self._alive: bool = True
        self._on_ref_error: RefErrorChoice = on_ref_error

        self.priority: int = priority

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        """Call the callback with `args`. Args will be spread when calling the func."""
        raise NotImplementedError()

    def dereference(self) -> Callable[..., _R] | None:
        """Return the original object, or None if dead."""
        raise NotImplementedError()

    def __call__(self, *args: Any, **kwds: Any) -> _R:
        func = self.dereference()
        if func is None:
            raise ReferenceError("callback is dead")
        if self._max_args is not None:
            args = args[: self._max_args]
        return func(*args, **kwds)

    def __eq__(self, other: object) -> bool:
        # sourcery skip: swap-if-expression
        if isinstance(other, WeakCallback):
            return self._key == other._key
        return NotImplemented

    def _try_ref(
        self,
        obj: _T,
        finalize: Callable[[WeakCallback], Any] | None = None,
    ) -> Callable[[], _T | None]:
        _cb = None if finalize is None else _kill_and_finalize(self, finalize)
        try:
            return weakref.ref(obj, _cb)
        except TypeError:
            if self._on_ref_error == "raise":
                raise
            if self._on_ref_error == "warn":
                safe_repr = object.__repr__(obj)
                warn(
                    f"failed to create weakref for {safe_repr}, returning strong ref",
                    stacklevel=2,
                )

            def _strong_ref() -> _T:
                return obj

            return _strong_ref

    def slot_repr(self) -> str:
        return f"{self._obj_module}.{self._obj_qualname}"

    @staticmethod
    def object_key(obj: Any) -> str:
        """Return a unique key for an object.

        This includes information about the object's type, module, and id. It has
        considerations for bound methods (which would otherwise have a different id
        for each instance).
        """
        if hasattr(obj, "__self__"):
            # bound method ... don't take the id of the bound method itself.
            obj_id = id(obj.__self__)
            owner_cls = type(obj.__self__)
            type_name = getattr(owner_cls, "__name__", None) or ""
            module = getattr(owner_cls, "__module__", None) or ""
            method_name = getattr(obj, "__name__", None) or ""
            obj_name = f"{type_name}.{method_name}"
        else:
            obj_id = id(obj)
            module = getattr(obj, "__module__", None) or ""
            obj_name = getattr(obj, "__name__", None) or ""
        return f"{module}:{obj_name}@{hex(obj_id)}"

    @staticmethod
    def object_repr(obj: Any) -> str:
        """Return a human-readable repr for obj."""
        module = getattr(obj, "__module__", "")
        if hasattr(obj, "__self__"):
            # bound method ... don't take the id of the bound method itself.
            owner_cls = type(obj.__self__)
            module = getattr(owner_cls, "__module__", None) or ""
            method_name = getattr(obj, "__name__", None) or ""
            if module == "builtins":
                return method_name
            type_qname = getattr(owner_cls, "__qualname__", "")
            return f"{module}.{type_qname}.{method_name}"
        elif getattr(obj, "__qualname__", ""):
            return f"{module}.{obj.__qualname__}"
        elif getattr(type(obj), "__qualname__", ""):
            return f"{module}.{type(obj).__qualname__}"
        # this line was hit in py3.7, but not afterwards.
        # retained as a fallback, but not covered by tests.
        return repr(obj)  # pragma: no cover

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} on {self._object_repr}>"  # pragma: no cover


def _kill_and_finalize(
    wcb: WeakCallback, finalize: Callable[[WeakCallback], Any]
) -> Callable[[weakref.ReferenceType], None]:
    def _cb(_: weakref.ReferenceType) -> None:
        if wcb._alive:
            wcb._alive = False
            finalize(wcb)

    return _cb


@mypyc_attr(serializable=True)
class StrongFunction(WeakCallback):
    """Wrapper around a strong function reference."""

    _f: Callable

    def __init__(
        self,
        obj: Callable,
        max_args: int | None = None,
        args: tuple[Any, ...] = (),
        kwargs: dict[str, Any] | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
    ) -> None:
        super().__init__(obj, max_args, on_ref_error, priority)
        self._f = obj
        self._args = args
        self._kwargs = kwargs or {}

        if args:
            self._object_repr = f"{self._object_repr}{(*args,)!r}".replace(")", " ...)")

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        if self._max_args is not None:
            args = args[: self._max_args]
        self._f(*self._args, *args, **self._kwargs)

    def dereference(self) -> Callable:
        if self._args or self._kwargs:
            return partial(self._f, *self._args, **self._kwargs)
        return self._f

    def __getstate__(self) -> dict[str, Any]:
        atr = ("_key", "_max_args", "_alive", "_on_ref_error", "_f", "_args", "_kwargs")
        return {k: getattr(self, k) for k in atr}

    def __setstate__(self, state: dict) -> None:
        for k, v in state.items():
            setattr(self, k, v)


class WeakFunction(WeakCallback):
    """Wrapper around a weak function reference."""

    def __init__(
        self,
        obj: Callable,
        max_args: int | None = None,
        args: tuple[Any, ...] = (),
        kwargs: dict[str, Any] | None = None,
        finalize: Callable | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
    ) -> None:
        super().__init__(obj, max_args, on_ref_error, priority)
        self._f = self._try_ref(obj, finalize)
        self._args = args
        self._kwargs = kwargs or {}

        if args:
            self._object_repr = f"{self._object_repr}{(*args,)!r}".replace(")", " ...)")

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        f = self._f()
        if f is None:
            raise ReferenceError("weakly-referenced object no longer exists")
        if self._max_args is not None:
            args = args[: self._max_args]
        f(*self._args, *args, **self._kwargs)

    def dereference(self) -> Callable | None:
        f = self._f()
        if f is None:
            return None
        if self._args or self._kwargs:
            return partial(f, *self._args, **self._kwargs)
        return f


class WeakMethod(WeakCallback):
    """Wrapper around a method bound to a weakly-referenced object.

    Bound methods have a `__self__` attribute that holds a strong reference to the
    object they are bound to and a `__func__` attribute that holds a reference
    to the function that implements the method (on the class level)

    When `cb` is called here, it dereferences the two, and calls:
    `obj.__func__(obj.__self__, *args, **kwargs)`
    """

    def __init__(
        self,
        obj: MethodType,
        max_args: int | None = None,
        args: tuple[Any, ...] = (),
        kwargs: dict[str, Any] | None = None,
        finalize: Callable | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
    ) -> None:
        super().__init__(obj, max_args, on_ref_error, priority)
        self._obj_ref = self._try_ref(obj.__self__, finalize)
        self._func_ref = self._try_ref(obj.__func__, finalize)
        self._args = args
        self._kwargs = kwargs or {}
        if args:
            self._object_repr = f"{self._object_repr}{(*args,)!r}".replace(")", " ...)")

    def slot_repr(self) -> str:
        obj = self._obj_ref()
        func_name = getattr(self._func_ref(), "__name__", "<method>")
        return f"{self._obj_module}.{obj.__class__.__qualname__}.{func_name}"

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        obj = self._obj_ref()
        func = self._func_ref()
        if obj is None or func is None:
            raise ReferenceError("weakly-referenced object no longer exists")

        if self._max_args is not None:
            args = args[: self._max_args]
        func(obj, *self._args, *args, **self._kwargs)

    def dereference(self) -> MethodType | partial | None:
        obj = self._obj_ref()
        func = self._func_ref()
        if obj is None or func is None:
            return None
        method = cast("MethodType", func.__get__(obj))
        if self._args or self._kwargs:
            return partial(method, *self._args, **self._kwargs)
        return method


class WeakBuiltin(WeakCallback):
    """Wrapper around a c-based method on a weakly-referenced object.

    Builtin/extension methods do have a `__self__` attribute (the object to which they
    are bound), but don't have a __func__ attribute, so we need to store the name of the
    method and look it up on the object when the callback is called.

    When `cb` is called here, it dereferences the object, and calls:
    `getattr(obj.__self__, obj.__name__)(*args, **kwargs)`
    """

    def __init__(
        self,
        obj: MethodWrapperType | BuiltinMethodType,
        max_args: int | None = None,
        args: tuple[Any, ...] = (),
        finalize: Callable | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
    ) -> None:
        super().__init__(obj, max_args, on_ref_error, priority)
        self._obj_ref = self._try_ref(obj.__self__, finalize)
        self._func_name = obj.__name__
        self._args = args
        if args:
            self._object_repr = f"{self._object_repr}{(*args,)!r}".replace(")", " ...)")

    def slot_repr(self) -> str:
        obj = self._obj_ref()
        return f"{obj.__class__.__qualname__}.{self._func_name}"

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        func = getattr(self._obj_ref(), self._func_name, None)
        if func is None:
            raise ReferenceError("weakly-referenced object no longer exists")
        if self._max_args is None:
            func(*self._args, *args)
        else:
            func(*self._args, *args[: self._max_args])

    def dereference(self) -> MethodWrapperType | BuiltinMethodType | None:
        return getattr(self._obj_ref(), self._func_name, None)


class WeakSetattr(WeakCallback):
    """Caller to set an attribute on a weakly-referenced object."""

    def __init__(
        self,
        obj: object,
        attr: str,
        max_args: int | None = None,
        finalize: Callable | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
    ) -> None:
        super().__init__(obj, max_args, on_ref_error, priority)
        self._key += f".__setattr__({attr!r})"
        self._obj_ref = self._try_ref(obj, finalize)
        self._attr = attr
        self._object_repr += f".__setattr__({attr!r}, ...)"

    def slot_repr(self) -> str:
        obj = self._obj_ref()
        return f"setattr({obj.__class__.__qualname__}, {self._attr!r}, ...)"

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        obj = self._obj_ref()
        if obj is None:
            raise ReferenceError("weakly-referenced object no longer exists")
        if self._max_args is not None:
            args = args[: self._max_args]
        setattr(obj, self._attr, args[0] if len(args) == 1 else args)

    def dereference(self) -> partial | None:
        obj = self._obj_ref()
        return None if obj is None else partial(setattr, obj, self._attr)


class SupportsSetitem(Protocol):
    def __setitem__(self, key: Any, value: Any) -> None: ...


class WeakSetitem(WeakCallback):
    """Caller to call __setitem__ on a weakly-referenced object."""

    def __init__(
        self,
        obj: SupportsSetitem,
        key: Any,
        max_args: int | None = None,
        finalize: Callable | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
    ) -> None:
        super().__init__(obj, max_args, on_ref_error, priority)
        self._key += f".__setitem__({key!r})"
        self._obj_ref = self._try_ref(obj, finalize)
        self._itemkey = key
        self._object_repr += f".__setitem__({key!r}, ...)"

    def slot_repr(self) -> str:
        obj = self._obj_ref()
        return f"{obj.__class__.__qualname__}.__setitem__({self._itemkey!r}, ...)"

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        obj = self._obj_ref()
        if obj is None:
            raise ReferenceError("weakly-referenced object no longer exists")
        if self._max_args is not None:
            args = args[: self._max_args]
        obj[self._itemkey] = args[0] if len(args) == 1 else args

    def dereference(self) -> partial | None:
        obj = self._obj_ref()
        return None if obj is None else partial(obj.__setitem__, self._itemkey)


# --------------------------- Coroutines ---------------------------


class WeakCoroutineFunction(WeakFunction):
    def cb(self, args: tuple[Any, ...] = ()) -> None:
        if self._f() is None:
            raise ReferenceError("weakly-referenced object no longer exists")
        if self._max_args is not None:
            args = args[: self._max_args]

        cast("_AsyncBackend", get_async_backend()).put((self, args))


class StrongCoroutineFunction(StrongFunction):
    """Wrapper around a strong coroutine function reference."""

    def cb(self, args: tuple[Any, ...] = ()) -> None:
        if self._max_args is not None:
            args = args[: self._max_args]

        cast("_AsyncBackend", get_async_backend()).put((self, args))


class WeakCoroutineMethod(WeakMethod):
    def cb(self, args: tuple[Any, ...] = ()) -> None:
        if self._obj_ref() is None or self._func_ref() is None:
            raise ReferenceError("weakly-referenced object no longer exists")

        if self._max_args is not None:
            args = args[: self._max_args]

        cast("_AsyncBackend", get_async_backend()).put((self, args))



from collections.abc import Callable
from functools import partial
from typing import Any, ClassVar, Generic, TypeVar
from weakref import finalize

try:
    from wrapt import ObjectProxy
except ImportError as e:
    raise type(e)(
        f"{e}. Please `pip install psygnal[proxy]` to use EventedObjectProxies"
    ) from e

from psygnal._group import SignalGroup
from psygnal._signal import Signal

T = TypeVar("T")
_UNSET = object()


class ProxyEvents(SignalGroup):
    """Events emitted by `EventedObjectProxy` and `EventedCallableObjectProxy`."""

    attribute_set = Signal(str, object)
    """Emitted when an attribute is set."""
    attribute_deleted = Signal(str)
    """Emitted when an attribute is deleted."""
    item_set = Signal(object, object)
    """Emitted when an item is set."""
    item_deleted = Signal(object)
    """Emitted when an item is deleted."""
    in_place = Signal(str, object)
    """Emitted when an in-place operation is performed."""


class CallableProxyEvents(ProxyEvents):
    """Events emitted by `EventedCallableObjectProxy`."""

    called = Signal(tuple, dict)
    """Emitted when the object is called."""


# we're using a cache instead of setting the events object directly on the proxy
# because when wrapt is compiled as a C extensions, the ObjectProxy is not allowed
# to add any new attributes.
_OBJ_CACHE: dict[int, ProxyEvents] = {}


class EventedObjectProxy(ObjectProxy, Generic[T]):
    """Create a proxy of `target` that includes an `events` [psygnal.SignalGroup][].

    Provides an "evented" subclasses of
    [`wrapt.ObjectProxy`](https://wrapt.readthedocs.io/en/latest/wrappers.html#object-proxy)

    !!! important

        This class requires `wrapt` to be installed.
        You can install directly (`pip install wrapt`) or by using the psygnal
        extra: `pip install psygnal[proxy]`

    Signals will be emitted whenever an attribute is set or deleted, or
    (if the object implements `__getitem__`) whenever an item is set or deleted.
    If the object supports in-place modification (i.e. any of the `__i{}__` magic
    methods), then an `in_place` event is emitted (with the name of the method)
    whenever any of them are used.

    The events available at target.events include:

    - `attribute_set`: `Signal(str, object)`
    - `attribute_deleted`: `Signal(str)`
    - `item_set`: `Signal(object, object)`
    - `item_deleted`: `Signal(object)`
    - `in_place`: `Signal(str, object)`

    !!! warning "Experimental"

        This object is experimental! They may affect the behavior of
        the wrapped object in unanticipated ways.  Please consult
        the [wrapt documentation](https://wrapt.readthedocs.io/en/latest/wrappers.html)
        for details on how the Object Proxy works.

    Parameters
    ----------
    target : Any
        An object to wrap
    """

    _psygnal_group_: ClassVar[str] = "events"

    def __init__(self, target: Any):
        super().__init__(target)

    @property
    def events(self) -> ProxyEvents:  # pragma: no cover # unclear why
        """`SignalGroup` containing events for this object proxy."""
        obj_id = id(self)
        if obj_id not in _OBJ_CACHE:
            _OBJ_CACHE[obj_id] = ProxyEvents()
            finalize(self, partial(_OBJ_CACHE.pop, obj_id, None))
        return _OBJ_CACHE[obj_id]

    def __setattr__(self, name: str, value: None) -> None:
        before = getattr(self, name, _UNSET)
        super().__setattr__(name, value)
        if before is not (after := getattr(self, name, _UNSET)):
            self.events.attribute_set(name, after)

    def __delattr__(self, name: str) -> None:
        super().__delattr__(name)
        self.events.attribute_deleted(name)

    def __setitem__(self, key: Any, value: Any) -> None:
        before = self[key]
        super().__setitem__(key, value)
        if before is not (after := self[key]):
            self.events.item_set(key, after)

    def __delitem__(self, key: Any) -> None:
        super().__delitem__(key)
        self.events.item_deleted(key)

    def __repr__(self) -> str:
        return repr(self.__wrapped__)

    def __dir__(self) -> list[str]:
        return [*dir(self.__wrapped__), "events"]

    def __iadd__(self, other: Any) -> T:
        self.events.in_place("add", other)
        return super().__iadd__(other)  # type: ignore

    def __isub__(self, other: Any) -> T:
        self.events.in_place("sub", other)
        return super().__isub__(other)  # type: ignore

    def __imul__(self, other: Any) -> T:
        self.events.in_place("mul", other)
        return super().__imul__(other)  # type: ignore

    def __imatmul__(self, other: Any) -> T:
        self.events.in_place("matmul", other)
        self.__wrapped__ @= other  # not in wrapt  # type: ignore
        return self

    def __itruediv__(self, other: Any) -> T:
        self.events.in_place("truediv", other)
        return super().__itruediv__(other)  # type: ignore

    def __ifloordiv__(self, other: Any) -> T:
        self.events.in_place("floordiv", other)
        return super().__ifloordiv__(other)  # type: ignore

    def __imod__(self, other: Any) -> T:
        self.events.in_place("mod", other)
        return super().__imod__(other)  # type: ignore

    def __ipow__(self, other: Any) -> T:
        self.events.in_place("pow", other)
        return super().__ipow__(other)  # type: ignore

    def __ilshift__(self, other: Any) -> T:
        self.events.in_place("lshift", other)
        return super().__ilshift__(other)  # type: ignore

    def __irshift__(self, other: Any) -> T:
        self.events.in_place("rshift", other)
        return super().__irshift__(other)  # type: ignore

    def __iand__(self, other: Any) -> T:
        self.events.in_place("and", other)
        return super().__iand__(other)  # type: ignore

    def __ixor__(self, other: Any) -> T:
        self.events.in_place("xor", other)
        return super().__ixor__(other)  # type: ignore

    def __ior__(self, other: Any) -> T:
        self.events.in_place("or", other)
        return super().__ior__(other)  # type: ignore


class EventedCallableObjectProxy(EventedObjectProxy):
    """Create a proxy of `target` that includes an `events` [psygnal.SignalGroup][].

    `target` must be callable.

    !!! important

        This class requires `wrapt` to be installed.
        You can install directly (`pip install wrapt`) or by using the psygnal
        extra: `pip install psygnal[proxy]`

    Signals will be emitted whenever an attribute is set or deleted, or
    (if the object implements `__getitem__`) whenever an item is set or deleted.
    If the object supports in-place modification (i.e. any of the `__i{}__` magic
    methods), then an `in_place` event is emitted (with the name of the method)
    whenever any of them are used.  Lastly, if the item is called, a `called`
    event is emitted with the (args, kwargs) used in the call.

    The events available at `target.events` include:

    - `attribute_set`: `Signal(str, object)`
    - `attribute_deleted`: `Signal(str)`
    - `item_set`: `Signal(object, object)`
    - `item_deleted`: `Signal(object)`
    - `in_place`: `Signal(str, object)`
    - `called`: `Signal(tuple, dict)`

    Parameters
    ----------
    target : Callable
        An callable object to wrap

    """

    def __init__(self, target: Callable):
        super().__init__(target)

    @property
    def events(self) -> CallableProxyEvents:  # pragma: no cover # unclear why
        """`SignalGroup` containing events for this object proxy."""
        obj_id = id(self)
        if obj_id not in _OBJ_CACHE:
            _OBJ_CACHE[obj_id] = CallableProxyEvents()
            finalize(self, partial(_OBJ_CACHE.pop, obj_id, None))
        return _OBJ_CACHE[obj_id]  # type: ignore

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Call the wrapped object and emit a `called` signal."""
        self.events.called(args, kwargs)
        return self.__wrapped__(*args, **kwargs)



"""Containers backed by psygnal events.

These classes provide "evented" versions of mutable python containers.
They each have an `events` attribute (`SignalGroup`) that has a variety of
signals that will emit whenever the container is mutated.  See
[Container SignalGroups](#container-signalgroups) for the corresponding
container type for details on the available signals.
"""

from typing import TYPE_CHECKING, Any

from ._evented_dict import DictEvents, EventedDict
from ._evented_list import EventedList, ListEvents
from ._evented_set import EventedOrderedSet, EventedSet, OrderedSet, SetEvents
from ._selectable_evented_list import SelectableEventedList
from ._selection import Selection

if TYPE_CHECKING:
    from ._evented_proxy import (
        CallableProxyEvents,
        EventedCallableObjectProxy,
        EventedObjectProxy,
        ProxyEvents,
    )

__all__ = [
    "CallableProxyEvents",
    "DictEvents",
    "EventedCallableObjectProxy",
    "EventedDict",
    "EventedList",
    "EventedObjectProxy",
    "EventedOrderedSet",
    "EventedSet",
    "ListEvents",
    "OrderedSet",
    "ProxyEvents",
    "SelectableEventedList",
    "Selection",
    "SetEvents",
]


def __getattr__(name: str) -> Any:  # pragma: no cover
    if name == "EventedObjectProxy":
        from ._evented_proxy import EventedObjectProxy

        return EventedObjectProxy
    if name == "EventedCallableObjectProxy":
        from ._evented_proxy import EventedCallableObjectProxy

        return EventedCallableObjectProxy
    if name == "CallableProxyEvents":
        from ._evented_proxy import CallableProxyEvents

        return CallableProxyEvents
    if name == "ProxyEvents":
        from ._evented_proxy import ProxyEvents

        return ProxyEvents
    raise AttributeError(  # pragma: no cover
        f"module {__name__!r} has no attribute {name!r}"
    )



"""MutableSequence with a selection model."""

from collections.abc import Iterable
from typing import Any, TypeVar

from ._evented_list import EventedList, ListEvents
from ._selection import Selectable

_T = TypeVar("_T")


class SelectableEventedList(Selectable[_T], EventedList[_T]):
    """`EventedList` subclass with a built in selection model.

    In addition to all `EventedList` properties, this class also has a `selection`
    attribute that manages a set of selected items in the list.

    Parameters
    ----------
    data : iterable, optional
        Elements to initialize the list with.
    hashable : bool
        Whether the list should be hashable as id(self). By default `True`.
    child_events: bool
        Whether to re-emit events from emitted from evented items in the list
        (i.e. items that have SignalInstances). If `True`, child events can be connected
        at `EventedList.events.child_event`. By default, `False`.

    Attributes
    ----------
    events : ListEvents
        SignalGroup that with events related to list mutation.  (see ListEvents)
    selection : Selection
        An evented set containing the currently selected items, along with an `active`
        and `current` item.  (See `Selection`)
    """

    events: ListEvents  # pragma: no cover

    def __init__(
        self,
        data: Iterable[_T] = (),
        *,
        hashable: bool = True,
        child_events: bool = False,
    ):
        self._activate_on_insert: bool = True
        super().__init__(data=data, hashable=hashable, child_events=child_events)
        self.events.removed.connect(self._on_item_removed)

    def _on_item_removed(self, idx: int, obj: Any) -> None:
        self.selection.discard(obj)

    def insert(self, index: int, value: _T) -> None:
        """Insert item(s) into the list and update the selection."""
        super().insert(index, value)
        if self._activate_on_insert:
            self.selection.active = value

    def select_all(self) -> None:
        """Select all items in the list."""
        self.selection.update(self)

    def deselect_all(self) -> None:
        """Deselect all items in the list."""
        self.selection.clear()

    def select_next(
        self, step: int = 1, expand_selection: bool = False, wraparound: bool = False
    ) -> None:
        """Select the next item in the list.

        Parameters
        ----------
        step : int
            The step size to take when picking the next item, by default 1
        expand_selection : bool
            If True, will expand the selection to contain the both the current item and
            the next item, by default False
        wraparound : bool
            Whether to return to the beginning of the list of the end has been reached,
            by default False
        """
        if len(self) == 0:
            return
        elif not self.selection:
            idx = -1 if step > 0 else 0
        else:
            idx = self.index(self.selection._current) + step
        idx_in_sequence = len(self) > idx >= 0
        if wraparound:
            idx = idx % len(self)
        elif not idx_in_sequence:
            idx = -1 if step > 0 else 0
        next_item = self[idx]
        if expand_selection:
            self.selection.add(next_item)
            self.selection._current = next_item
        else:
            self.selection.active = next_item

    def select_previous(
        self, expand_selection: bool = False, wraparound: bool = False
    ) -> None:
        """Select the previous item in the list."""
        self.select_next(
            step=-1, expand_selection=expand_selection, wraparound=wraparound
        )

    def remove_selected(self) -> tuple[_T, ...]:
        """Remove selected items from the list and the selection.

        Returns
        -------
        Tuple[_T, ...]
            The items that were removed.
        """
        selected_items = tuple(self.selection)
        idx = 0
        for item in list(self.selection):
            idx = self.index(item)
            self.remove(item)
        new_idx = max(0, idx - 1)
        if len(self) > new_idx:
            self.selection.add(self[new_idx])
        return selected_items



"""MutableSequence that emits events when altered.

Note For Developers
===================

Be cautious when re-implementing typical list-like methods here (e.g. extend,
pop, clear, etc...).  By not re-implementing those methods, we force ALL "CRUD"
(create, read, update, delete) operations to go through a few key methods
defined by the abc.MutableSequence interface, where we can emit the necessary
events.

Specifically:

- `insert` = "create" : add a new item/index to the list
- `__getitem__` = "read" : get the value of an existing index
- `__setitem__` = "update" : update the value of an existing index
- `__delitem__` = "delete" : remove an existing index from the list

All of the additional list-like methods are provided by the MutableSequence
interface, and call one of those 4 methods.  So if you override a method, you
MUST make sure that all the appropriate events are emitted.  (Tests should
cover this in test_evented_list.py)
"""

from __future__ import annotations

from collections.abc import Iterable, Mapping, MutableSequence
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    TypeAlias,
    TypeVar,
    cast,
    get_args,
    overload,
)

from psygnal._group import EmissionInfo, PathStep, SignalGroup
from psygnal._signal import Signal, SignalInstance
from psygnal.utils import iter_signal_instances

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler, SerializationInfo
    from typing_extensions import Self

_T = TypeVar("_T")
Index: TypeAlias = int | slice


class ListSignalInstance(SignalInstance):
    def _psygnal_relocate_info_(self, emission_info: EmissionInfo) -> EmissionInfo:
        """Relocate the emission info to the index being modified.

        (All signals on EventedList have the index as the first argument.)
        """
        if args := emission_info.args:
            return emission_info.insert_path(PathStep(index=args[0]))
        return emission_info


ListSignal = partial(Signal, signal_instance_class=ListSignalInstance)


class ListEvents(SignalGroup):
    """Events available on [EventedList][psygnal.containers.EventedList]."""

    inserting = ListSignal(int)
    """`(index)` emitted before an item is inserted at `index`"""
    inserted = ListSignal(int, object)
    """`(index, value)` emitted after `value` is inserted at `index`"""
    removing = ListSignal(int)
    """`(index)` emitted before an item is removed at `index`"""
    removed = ListSignal(int, object)
    """`(index, value)` emitted after `value` is removed at `index`"""
    moving = ListSignal(int, int)
    """`(index, new_index)` emitted before an item is moved from `index` to
    `new_index`"""
    moved = ListSignal(int, int, object)
    """`(index, new_index, value)` emitted after `value` is moved from
    `index` to `new_index`"""
    changed = ListSignal(object, object, object)
    """`(index_or_slice, old_value, value)` emitted when `index` is set from
    `old_value` to `value`"""
    reordered = Signal()
    """Emitted when the list is reordered (eg. moved/reversed)."""
    child_event = Signal(EmissionInfo)
    """`(EmissionInfo)` emitted when an object in the list emits an
    event. Note that the `EventedList` must be created with `child_events=True` in
    order for this to be emitted.
    """


class EventedList(MutableSequence[_T]):
    """Mutable Sequence that emits events when altered.

    This class is designed to behave exactly like the builtin `list`, but
    will emit events before and after all mutations (insertion, removal,
    setting, and moving).

    Parameters
    ----------
    data : iterable, optional
        Elements to initialize the list with.
    hashable : bool
        Whether the list should be hashable as id(self). By default `True`.
    child_events: bool
        Whether to re-emit events from emitted from evented items in the list
        (i.e. items that have SignalInstances). If `True`, child events can be connected
        at `EventedList.events.child_event`. By default, `True`.

    Attributes
    ----------
    events : ListEvents
        SignalGroup that with events related to list mutation.  (see ListEvents)
    """

    events: ListEvents  # pragma: no cover
    _psygnal_group_: ClassVar[str] = "events"

    def __init__(
        self,
        data: Iterable[_T] = (),
        *,
        hashable: bool = True,
        child_events: bool = True,
    ):
        super().__init__()
        self._data: list[_T] = []
        self._hashable = hashable
        self._child_events = child_events
        self.events = ListEvents(instance=self)
        self.extend(data)

    # WAIT!! ... Read the module docstring before reimplement these methods
    # def append(self, item): ...
    # def clear(self): ...
    # def pop(self, index=-1): ...
    # def extend(self, value: Iterable[_T]): ...
    # def remove(self, value: Any): ...

    def insert(self, index: int, value: _T) -> None:
        """Insert `value` before index."""
        _value = self._pre_insert(value)
        self.events.inserting.emit(index)
        self._data.insert(index, _value)
        self.events.inserted.emit(index, value)
        self._post_insert(value)

    @overload
    def __getitem__(self, key: int) -> _T: ...

    @overload
    def __getitem__(self, key: slice) -> Self: ...

    def __getitem__(self, key: Index) -> _T | Self:
        """Return self[key]."""
        result = self._data[key]
        return self.__newlike__(result) if isinstance(result, list) else result

    @overload
    def __setitem__(self, key: int, value: _T) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[_T]) -> None: ...

    def __setitem__(self, key: Index, value: _T | Iterable[_T]) -> None:
        """Set self[key] to value."""
        old = self._data[key]
        if value is old:
            return

        # sourcery skip: hoist-similar-statement-from-if, hoist-statement-from-if
        if isinstance(key, slice):
            if not isinstance(value, Iterable):
                raise TypeError("Can only assign an iterable to slice")
            value = [self._pre_insert(v) for v in value]  # before we mutate the list
            self._data[key] = value
        else:
            value = self._pre_insert(cast("_T", value))
            self._data[key] = value

        self.events.changed.emit(key, old, value)

    def __delitem__(self, key: Index) -> None:
        """Delete self[key]."""
        # delete from the end
        for parent, index in sorted(self._delitem_indices(key), reverse=True):
            parent.events.removing.emit(index)
            parent._pre_remove(index)
            item = parent._data.pop(index)
            self.events.removed.emit(index, item)

    def _delitem_indices(self, key: Index) -> Iterable[tuple[EventedList[_T], int]]:
        # returning (self, int) allows subclasses to pass nested members
        if isinstance(key, int):
            yield (self, key if key >= 0 else key + len(self))
        elif isinstance(key, slice):
            yield from ((self, i) for i in range(*key.indices(len(self))))
        else:
            n = repr(type(key).__name__)
            raise TypeError(f"EventedList indices must be integers or slices, not {n}")

    def _pre_insert(self, value: _T) -> _T:
        """Validate and or modify values prior to inserted."""
        return value

    def _post_insert(self, new_item: _T) -> None:
        """Modify and or handle values after insertion."""
        if self._child_events:
            self._connect_child_emitters(new_item)

    def _pre_remove(self, index: int) -> None:
        """Modify and or handle values before removal."""
        if self._child_events:
            self._disconnect_child_emitters(self[index])

    def __newlike__(self, iterable: Iterable[_T]) -> Self:
        """Return new instance of same class."""
        return self.__class__(iterable)

    def copy(self) -> Self:
        """Return a shallow copy of the list."""
        return self.__newlike__(self)

    def __copy__(self) -> Self:
        return self.copy()

    def __add__(self, other: Iterable[_T]) -> Self:
        """Add other to self, return new object."""
        copy = self.copy()
        copy.extend(other)
        return copy

    def __iadd__(self, other: Iterable[_T]) -> Self:
        """Add other to self in place (self += other)."""
        self.extend(other)
        return self

    def __radd__(self, other: list) -> list:
        """Reflected add (other + self).  Cast self to list."""
        return other + list(self)

    def __len__(self) -> int:
        """Return len(self)."""
        return len(self._data)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"{type(self).__name__}({self._data})"

    def __eq__(self, other: Any) -> bool:
        """Return self==value."""
        return bool(self._data == other)

    def __hash__(self) -> int:
        """Return hash(self)."""
        # it's important to add this to allow this object to be hashable
        # given that we've also reimplemented __eq__
        if self._hashable:
            return id(self)
        name = self.__class__.__name__
        raise TypeError(
            f"unhashable type: {name!r}. "
            f"Create with {name}(..., hashable=True) if you need hashability"
        )

    def reverse(self, *, emit_individual_events: bool = False) -> None:
        """Reverse list *IN PLACE*."""
        if emit_individual_events:
            super().reverse()
        else:
            self._data.reverse()
        self.events.reordered.emit()

    def move(self, src_index: int, dest_index: int = 0) -> bool:
        """Insert object at `src_index` before `dest_index`.

        Both indices refer to the list prior to any object removal
        (pre-move space).
        """
        if dest_index < 0:
            dest_index += len(self) + 1
        if dest_index in (src_index, src_index + 1):
            # this is a no-op
            return False

        self.events.moving.emit(src_index, dest_index)
        item = self._data.pop(src_index)
        if dest_index > src_index:
            dest_index -= 1
        self._data.insert(dest_index, item)
        self.events.moved.emit(src_index, dest_index, item)
        self.events.reordered.emit()
        return True

    def move_multiple(self, sources: Iterable[Index], dest_index: int = 0) -> int:
        """Move a batch of `sources` indices, to a single destination.

        Note, if `dest_index` is higher than any of the `sources`, then
        the resulting position of the moved objects after the move operation
        is complete will be lower than `dest_index`.

        Parameters
        ----------
        sources : Iterable[Union[int, slice]]
            A sequence of indices
        dest_index : int, optional
            The destination index.  All sources will be inserted before this
            index (in pre-move space), by default 0... which has the effect of
            "bringing to front" everything in `sources`, or acting as a
            "reorder" method if `sources` contains all indices.

        Returns
        -------
        int
            The number of successful move operations completed.

        Raises
        ------
        TypeError
            If the destination index is a slice, or any of the source indices
            are not `int` or `slice`.
        """
        # calling list here makes sure that there are no index errors up front
        move_plan = list(self._move_plan(sources, dest_index))

        # don't assume index adjacency ... so move objects one at a time
        # this *could* be simplified with an intermediate list ... but this way
        # allows any views (such as QtViews) to update themselves more easily.
        # If this needs to be changed in the future for performance reasons,
        # then the associated QtListView will need to changed from using
        # `beginMoveRows` & `endMoveRows` to using `layoutAboutToBeChanged` &
        # `layoutChanged` while *manually* updating model indices with
        # `changePersistentIndexList`.  That becomes much harder to do with
        # nested tree-like models.
        with self.events.reordered.blocked():
            for src, dest in move_plan:
                self.move(src, dest)

        self.events.reordered.emit()
        return len(move_plan)

    def _move_plan(
        self, sources: Iterable[Index], dest_index: int
    ) -> Iterable[tuple[int, int]]:
        """Yield prepared indices for a multi-move.

        Given a set of `sources` from anywhere in the list,
        and a single `dest_index`, this function computes and yields
        `(from_index, to_index)` tuples that can be used sequentially in
        single move operations.  It keeps track of what has moved where and
        updates the source and destination indices to reflect the model at each
        point in the process.

        This is useful for a drag-drop operation with a QtModel/View.

        Parameters
        ----------
        sources : Iterable[tuple[int, ...]]
            An iterable of tuple[int] that should be moved to `dest_index`.
        dest_index : Tuple[int]
            The destination for sources.
        """
        if isinstance(dest_index, slice):
            raise TypeError("Destination index may not be a slice")  # pragma: no cover

        to_move: list[int] = []
        for idx in sources:
            if isinstance(idx, slice):
                to_move.extend(list(range(*idx.indices(len(self)))))
            elif isinstance(idx, int):
                to_move.append(idx)
            else:
                raise TypeError(
                    "Can only move integer or slice indices"
                )  # pragma: no cover

        to_move = list(dict.fromkeys(to_move))

        if dest_index < 0:
            dest_index += len(self) + 1

        d_inc = 0
        popped: list[int] = []
        for i, src in enumerate(to_move):
            if src != dest_index:
                # we need to decrement the src_i by 1 for each time we have
                # previously pulled items out from in front of the src_i
                src -= sum(x <= src for x in popped)
                # if source is past the insertion point, increment src for each
                # previous insertion
                if src >= dest_index:
                    src += i
                yield src, dest_index + d_inc

            popped.append(src)
            # if the item moved up, increment the destination index
            if dest_index <= src:
                d_inc += 1

    def _connect_child_emitters(self, child: _T) -> None:
        """Connect all events from the child to be reemitted."""
        for emitter in iter_signal_instances(child):
            emitter.connect(self._reemit_child_event)

    def _disconnect_child_emitters(self, child: _T) -> None:
        """Disconnect all events from the child from the reemitter."""
        for emitter in iter_signal_instances(child):
            emitter.disconnect(self._reemit_child_event)

    def _reemit_child_event(self, *args: Any) -> None:
        """Re-emit event from child with index."""
        emitter = Signal.current_emitter()
        if emitter is None:
            return  # pragma: no cover
        obj = emitter.instance
        try:
            idx = self.index(obj)
        except ValueError:  # pragma: no cover
            return

        if args and isinstance(args[0], EmissionInfo):
            child_info = EmissionInfo(
                signal=args[0].signal,
                args=args[0].args,
                path=(PathStep(index=idx), *args[0].path),
            )
        else:
            child_info = EmissionInfo(
                signal=emitter,
                args=args,
                path=(PathStep(index=idx), PathStep(attr=emitter.name)),
            )

        self.events.child_event.emit(child_info)

    # PYDANTIC SUPPORT

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> Mapping[str, Any]:
        """Return the Pydantic core schema for this object."""
        from pydantic_core import core_schema

        def _serialize(obj: EventedList[_T], info: SerializationInfo, /) -> Any:
            if info.mode_is_json():
                return obj._data
            return cls(obj._data)

        item_type = args[0] if (args := get_args(source_type)) else Any
        items_schema = handler.generate_schema(item_type)
        list_schema = core_schema.list_schema(items_schema=items_schema)
        return core_schema.no_info_after_validator_function(
            function=cls,
            schema=list_schema,
            json_schema_input_schema=list_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                _serialize,
                info_arg=True,
            ),
        )



from __future__ import annotations

import contextlib
import dataclasses
import sys
from types import GenericAlias
from typing import TYPE_CHECKING, Any, Protocol, cast, overload

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import TypeGuard

    import attrs
    import msgspec
    from pydantic import BaseModel


class _DataclassParams(Protocol):
    init: bool
    repr: bool
    eq: bool
    order: bool
    unsafe_hash: bool
    frozen: bool


class AttrsType:
    __attrs_attrs__: tuple[attrs.Attribute, ...]


_DATACLASS_PARAMS = "__dataclass_params__"
with contextlib.suppress(ImportError):
    from dataclasses import _DATACLASS_PARAMS  # type: ignore
_DATACLASS_FIELDS = "__dataclass_fields__"
with contextlib.suppress(ImportError):
    from dataclasses import _DATACLASS_FIELDS  # type: ignore


class DataClassType:
    __dataclass_params__: _DataclassParams
    __dataclass_fields__: dict[str, dataclasses.Field]


@overload
def is_dataclass(obj: type) -> TypeGuard[type[DataClassType]]: ...


@overload
def is_dataclass(obj: object) -> TypeGuard[DataClassType]: ...


def is_dataclass(obj: object) -> TypeGuard[DataClassType]:
    """Return True if the object is a dataclass."""
    cls = (
        obj
        if isinstance(obj, type) and not isinstance(obj, GenericAlias)
        else type(obj)
    )
    return hasattr(cls, _DATACLASS_FIELDS)


@overload
def is_attrs_class(obj: type) -> TypeGuard[type[AttrsType]]: ...


@overload
def is_attrs_class(obj: object) -> TypeGuard[AttrsType]: ...


def is_attrs_class(obj: object) -> TypeGuard[type[AttrsType]]:
    """Return True if the class is an attrs class."""
    attr = sys.modules.get("attr", None)
    cls = obj if isinstance(obj, type) else type(obj)
    return attr.has(cls) if attr is not None else False


@overload
def is_pydantic_model(obj: type) -> TypeGuard[type[BaseModel]]: ...


@overload
def is_pydantic_model(obj: object) -> TypeGuard[BaseModel]: ...


def is_pydantic_model(obj: object) -> TypeGuard[BaseModel]:
    """Return True if the class is a pydantic BaseModel."""
    pydantic = sys.modules.get("pydantic", None)
    cls = obj if isinstance(obj, type) else type(obj)
    return pydantic is not None and issubclass(cls, pydantic.BaseModel)


@overload
def is_msgspec_struct(obj: type) -> TypeGuard[type[msgspec.Struct]]: ...


@overload
def is_msgspec_struct(obj: object) -> TypeGuard[msgspec.Struct]: ...


def is_msgspec_struct(obj: object) -> TypeGuard[msgspec.Struct]:
    """Return True if the class is a `msgspec.Struct`."""
    msgspec = sys.modules.get("msgspec", None)
    cls = obj if isinstance(obj, type) else type(obj)
    return msgspec is not None and issubclass(cls, msgspec.Struct)


def is_frozen(obj: Any) -> bool:
    """Return True if the object is frozen."""
    # sourcery skip: reintroduce-else
    cls = obj if isinstance(obj, type) else type(obj)

    params = cast("_DataclassParams | None", getattr(cls, _DATACLASS_PARAMS, None))
    if params is not None:
        return params.frozen

    # pydantic
    cfg = getattr(cls, "__config__", None)
    if cfg is not None and getattr(cfg, "allow_mutation", None) is False:
        return True

    # pydantic v2
    cfg = getattr(cls, "model_config", None)
    if cfg is not None and cfg.get("frozen"):
        return True

    # attrs
    if getattr(cls.__setattr__, "__name__", None) == "_frozen_setattrs":
        return True

    cfg = getattr(cls, "__struct_config__", None)
    if cfg is not None:  # pragma: no cover
        # this will be covered in msgspec > 0.13.1
        return bool(getattr(cfg, "frozen", False))

    return False


def iter_fields(
    cls: type, exclude_frozen: bool = True
) -> Iterator[tuple[str, type | None]]:
    """Iterate over all fields in the class, including inherited fields.

    This function recognizes dataclasses, attrs classes, msgspec Structs, and pydantic
    models.

    Parameters
    ----------
    cls : type
        The class to iterate over.
    exclude_frozen : bool, optional
        If True, frozen fields will be excluded. By default True.

    Yields
    ------
    tuple[str, type | None]
        The name and type of each field.
    """
    # generally opting for speed here over public API

    if (dclass_fields := getattr(cls, "__dataclass_fields__", None)) is not None:
        for d_field in dclass_fields.values():
            if d_field._field_type is dataclasses._FIELD:  # type: ignore [attr-defined]
                yield d_field.name, d_field.type
        return

    if is_pydantic_model(cls):
        if hasattr(cls, "model_fields"):
            for field_name, p_field in cls.model_fields.items():
                if not p_field.frozen or not exclude_frozen:
                    yield field_name, p_field.annotation
        else:
            for p_field in cls.__fields__.values():  # type: ignore [attr-defined]
                if p_field.field_info.allow_mutation or not exclude_frozen:
                    yield p_field.name, p_field.outer_type_
        return

    if (attrs_fields := getattr(cls, "__attrs_attrs__", None)) is not None:
        for a_field in attrs_fields:
            yield a_field.name, a_field.type
        return

    if is_msgspec_struct(cls):
        for m_field in cls.__struct_fields__:
            type_ = cls.__annotations__.get(m_field, None)
            yield m_field, type_
        return
